/* 
 * CalendarP.h - Private definitions for Calendar widget
 * 
 */

#ifndef _XtCalendarP_h
#define _XtCalendarP_h

/***********************************************************************
 *
 * Calendar Widget Private Data
 *
 ***********************************************************************/

#include "Calendar.h"
#include <X11/IntrinsicP.h>
#include <X11/Xmu/Converters.h>

/* New fields for the Calendar widget class record */

typedef struct {int foo;} CalendarClassPart;

/* Full class record declaration */
typedef struct _CalendarClassRec {
    CoreClassPart	core_class;
    CalendarClassPart	calendar_class;
} CalendarClassRec;

extern CalendarClassRec calendarClassRec;

/* New fields for the Calendar widget record */
typedef struct {
  XtCallbackList    clickProc;
  XtCallbackList    dateProc;
  XtCallbackList    dateExtProc;
  XtCallbackList    monthProc;
  XtCallbackList    yearProc;

  Pixel             foreground_pixel;
  Boolean           reverse_video;

  Boolean           showDays;
  Boolean           showMonth;
  Boolean           showYear;
  Boolean           drawGrid;
  XtJustify         dateJustify;

  GC		    normal_GC;
  GC                dashed_GC;
  GC                reverse_GC;
  GC                filled_GC[5];
  GC                selected_GC;

  XFontStruct       *font;
  XFontStruct       *dayFont;
  XFontStruct       *monthFont;
  XFontStruct       *hiliteFont;

  int               calendar_width,       /* following values computed */
                    calendar_height;      /* from width and height... */
  int               days_top;
  int               cells_top,
                    cells_left;
  Dimension         cell_BorderWidth;
  Dimension         cell_width,
                    cell_height;
  int               hStep, vStep,
                    hMax, vMax;
  int               monthLeft, monthRight, /* dependent on width, font */
                    yearLeft, yearRight;

  int               extended;
  int               select_start_x,
                    select_start_y,
                    select_end_x,
                    select_end_y;
  int               select_mode; /* ADDING or SUBING */

  /* in the case of weekly mode, >100 is needed, and it will clobber
     mapping, but that's ok because we don't need mapping. :) */
  int               selection[500];  /* width*height < 100; do dynamically? */
  int               mapping[100];    /* mappings needed to go from grid to */
  int               invmapping[100]; /* day of month and vice-versa */

  int               normal_ascent;
  int               day;        /* for week mode only */
  int               month;      /* for both month     */
  int               year;       /*  and week mode     */
  int               whatDay;    /* 1-7 for Sun - Sat  */
  char              strMY[30];  /* string containing Month, Year text */

  Boolean           mode;       /* 0 for month, 1 for week */
  int               numDays;    /* number of days to display in weekly mode */
  busy              *b;
  int               changed;

  Pixmap            bitsSelect, bits0, bits25, bits50, bits75, bits100;
} CalendarPart;


/****************************************************************
 *
 * Full instance record declaration
 *
 ****************************************************************/

typedef struct _CalendarRec {
  CorePart	core;
  CalendarPart	calendar;
} CalendarRec;

#define IN_BORDER -1

#define C_SELECT 0
#define C_UNSELECT 1

#define NOTHING 0
#define ADDING 1
#define SUBING 2

#define NOTIN 0
#define INMONTH 1
#define INYEAR 2

#endif _XtCalendarP_h
