#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <hesiod.h>
#include <X11/IntrinsicP.h>
#include <X11/Shell.h>
#include <Xm/Xm.h>
#include <Xm/DrawingA.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/Mu.h>
#include <Xm/PushB.h>
#include <Xm/ArrowB.h>
#include "MuMenu.h"
#include "structs.h"
#include "Calendar.h"

/* #define verbose */

extern Widget top[2]; /* to go into include... */
extern char *suffix();
extern char *month();
extern Widget week;

int poppedUp = 0, exposed = 0;

Arg args[10];
extern CalendarSelection displayed;
CalendarSelection sched;
Widget apptShell, apptForm, apptMenu, userList, name, user, freeTable;
Widget schedHeader, next, prev, userlistlab;

extern Widget mainShell, mainForm;

extern int numMonthChildren, numBusyChildren;
extern Widget monthChildren[], busyChildren[];

extern event *events;
busy busytimes;
int newbusy = 0;

#define MAXUSERS 50
int numusers = 0;
int listmap[MAXUSERS];

char *usernames[MAXUSERS];
busy *busies[MAXUSERS];
int ignore[MAXUSERS];
event *userevents[MAXUSERS];

extern Display *disp;
extern Window root;
extern Screen *screen;
extern screennum;
GC gc, dotgc;
GC gcs[5];
int black, white;
char errors[2000];

void eventsToBusy();
void redraw();
void addToUsers();
void updateGraph();

int drawgw, drawgh;
int drawlm = 60, drawtm = 14; /* left and top margins */

/* if ever called other than on startup, must free data structures */
void readpeople()
{
  FILE *f;
  char filename[50];
  char buffer[50];

  numusers = 0;

  XtSetArg( args[0], XmNselectedItems, NULL );
  XtSetArg( args[1], XmNselectedItemCount, 0 );
  XtSetArg( args[2], XmNitems, NULL );
  XtSetArg( args[3], XmNitemCount, 0 );
  XtSetValues( userList, args, 4 );

  sprintf( filename, "%s/.people", getenv( "HOME" ) );

  f = fopen( filename, "r" );
  if ( f == NULL )
    return;

  while ( !feof( f ) )
    {
      fscanf( f, "%s\n", buffer );
      addToUsers( buffer );
    }
  fclose( f );
}

void writepeople()
{
  FILE *f;
  char filename[50];
  int i;

  if ( !poppedUp )
    return;
  sprintf( filename, "%s/.people", getenv( "HOME" ) );

  f = fopen( filename, "w" );
  for ( i = 0; i < numusers; i++ )
    fprintf( f, "%s\n", usernames[listmap[i]] );

  fclose( f );
}

/*
 * For motif brain damage
 */
void MyXtUnmanageChildren( widgets, numWidgets )
Widget *widgets;
int numWidgets;
{
  int i;

  for (i = 0; i < numWidgets; i++ )
    XtUnmanageChild( widgets[i] );
}

void MyXtManageChildren( widgets, numWidgets )
Widget *widgets;
int numWidgets;
{
  int i;

  for (i = 0; i < numWidgets; i++ )
    XtManageChild( widgets[i] );
}

void appts( widget, closure, call_data )
Widget widget;
caddr_t closure, call_data;
{
  static int showing = 0;

  if ( showing )
    {
      showing = 0;

      MyXtUnmanageChildren(busyChildren, numBusyChildren);
      XtManageChildren(monthChildren, numMonthChildren);
      return;
    }

  showing = 1;
  if ( !poppedUp )
    {
      poppedUp = 1;
      readpeople();
    }

  XtUnmanageChildren(monthChildren, numMonthChildren);
  MyXtManageChildren(busyChildren, numBusyChildren);
}

void flush( widget, closure, call_data )
Widget widget;
caddr_t closure, call_data;
{
  int i;
  
  for ( i = 0; i < numusers; i++ )
    {
      ignore[i] = 0;
      if ( busies[i] != NULL )
	{
	  XtFree( busies[i] );
	  busies[i] = NULL;
	}
      if ( userevents[i] != NULL )
	freeEvents( &userevents[i] );
    }
  updateGraph();
}

char *appthelptext[3] = {
"   Type a username in the box labeled username and press return to add\n\
it to the list. The grid at the top is for showing times that people\n\
are busy. If you select (by clicking on) a username in the list, it\n\
will add their busy times to the ones already displayed in the grid.\n\
Similarly, clicking again will deselect the user and remove their busy\n\
times.\n\
\n\
   Non-white areas are shaded to indicate the number of people who\n\
have something scheduled at that time. If an area is black, that means\n\
four or more people have something scheduled then.\n\
\n\
   Thus, to find possible times to make appointments with a group of\n\
people (according to their calendars, anyway), select all of the\n\
people you want in the list, and the white areas leftover will be the\n\
open times of the group. At this time, you must do the legwork for\n\
actually making the appointment yourself.\n\
\n\
   Note that at this time, for someone to be able to see the times\n\
you are busy, they must have read access to your calendar file\n\
(.tc in the top of your home directory).",
"   The first time you select a given user, their calendar file is\n\
loaded in. Subsequent times, it uses the previously loaded file. Thus,\n\
if you selected someone at the beginning of the day, left the program\n\
running for a couple of hours, and came back, the data might be stale.\n\
The reread calendar files option will force the program to reload\n\
calendars so you're sure to have current information."
};

void apptHelp( widget, closure, call_data )
Widget widget;
int closure;
caddr_t call_data;
{
  MyHelp( 1, appthelptext[closure-1] );
}

void nextprevproc( widget, closure, call_data )
Widget widget;
int closure;
int call_data;
{
  offsetDays( closure, &sched.month, &sched.day, &sched.year );
/*  setHeader(); */
  updateGraph();
  redraw();
}

/* return the date of the first Sunday of the week */
firstOfWeek( month, day, year )
int *month, *day, *year;
{
  int c[42], i;

  cal( *month, *year, c );

  for ( i = 0; i < 42; i++ )
    if ( *day == c[i] )
      break; /* this had better not fail */

  i -= i%7; /* truncate to Sunday */
  if ( c[i] == 0 )
    {
      *month = *month - 1;
      if ( *month < 1 )
	{
	  *month = 12;
	  *year--;
	}
      cal( *month, *year, c );
      i = 35;
      while ( c[i] == 0 )
	i -= 7;
    }
  *day = c[i];
}

offsetDays( num, month, day, year )
int num, *month, *day, *year;
{
  int c[42], i, j, inc;
  int clue;

  inc = (num<0) ? -1 : 1;
  if ( num < 0 )
    num *= -1;

  cal( *month, *year, c );

  for ( i = 0; i < 42; i++ )
    if ( *day == c[i] )
      break; /* this had better not fail */

  for ( j = 0; j < num; j++ )
    {
      i += inc;

      clue = 0;
      if ( i < 0 || i > 41 )
	clue = 1;
      else
	if ( c[i] == 0 )
	  clue = 1;

      if ( clue )
	{
	  if ( inc == 1 )
	    {
	      (*month)++;
	      if ( *month > 12 )
		{
		  *month = 1;
		  (*year)++;
		}
	      cal( *month, *year, c );
	      i = i%7; /* this is safe; c[i] will be 1 */
	    }
	  else
	    {
	      (*month)--;
	      if ( *month < 1 )
		{
		  (*month) = 12;
		  (*year)--; /* bug - decrement to zero; lots of places */
		} /* insert big-bang graphic here :) */
	      cal( *month, *year, c );
	      i += 7; /* forces it non-negative so % will work */
	      i = i%7 + 35;
	      while ( c[i] == 0 )
		i -= 7;
	    }
	}
    }
  *day = c[i];
}

void eventsToBusy( events, bsy, month, day, year )
event *events;
busy *bsy;
int month, day, year;
{
  simpleEvent *s, *p;
  int c[42];
  int i, j, rolled = 0;
  int st, et, div;

  for ( i = 0; i < DAYS; i++ )
    for ( j = 0; j < HOURS*HDIVISIONS; j++ )
      bsy->when[i][j] = 0;

  cal( month, year, c );
  
  for ( i = 0; i < 42; i++ )
    if ( day == c[i] )
      break; /* this had better not fail */

  for (j = 0; j < DAYS; j++ )
    {
      if ( !rolled )
	{
	  if ( (day = c[i+j]) == 0 )
	    {
	      rolled = 1;
	      day = 1; /* 1752 */
	      month++;
	      if ( month > 12 )
		{
		  month = 1;
		  year++;
		}
	    }
	}
      else
	day++;

      if ( day != 0 )
	{
	  s = cvtEventToSimple( events, year, month, day );
	  p = s;

	  while ( p != NULL )
	    {
	      st = p->info->starttime.hour * 60 + p->info->starttime.minute;
	      et = p->info->endtime.hour * 60 + p->info->endtime.minute;
	      st = st / (60 / HDIVISIONS);
	      if ( et%(60/HDIVISIONS) == 0 )
		et -= (60/HDIVISIONS);
	      else
		et -= et%(60/HDIVISIONS);
	      et = et / (60/HDIVISIONS);

	      for ( div = st; div <= et; div++ )
		bsy->when[j][div] = 1;
	      p = p->next;
	    }
	  freeSimple( s );
	}
    }
}

static char *daysOfWeek[] = {
"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"
};

void redraw()
{
  XtSetArg( args[0], XmNbusy, &busytimes );
  XtSetArg( args[1], XmNday, sched.day );
  XtSetArg( args[2], XmNmonth, sched.month );
  XtSetArg( args[3], XmNyear, sched.year );
  XtSetArg( args[4], XmNchanged, ++newbusy );
  XtSetValues( week, args, 5 );
}

void addToUsers( name )
char *name;
{
  int comp, i, j, pos = 0, last = 1;
  char *stupidstr;

  for ( i = 0; i < numusers && last; i++ )
    {
      comp = strcasecmp( name, usernames[listmap[i]] );
      if ( comp == 0 )
	{
	  sprintf( errors, "User %s already in list.", name );
	  MyError( 1, errors );
	  return; /* sorta bad if user edits multiple duplicates into file */
	}
      else
	if ( comp < 0 ) /* name < usernames[i] */
	  {
	    last = 0;
	    pos = i + 1;
	    for ( j = numusers; j > i; j-- )
	      listmap[j] = listmap[j-1];
	    listmap[i] = numusers;
	    numusers++;
	    break;
	  }
    }
  if ( last )
    {
      listmap[numusers] = numusers;
      numusers++;
    }

  usernames[numusers-1] = XtNewString( name );
  busies[numusers-1] = NULL;
  userevents[numusers-1] = NULL;
  ignore[numusers-1] = 0;

  stupidstr = XmStringCreateLtoR( name,
				  XmSTRING_DEFAULT_CHARSET );

  XmListAddItem( userList, stupidstr, pos );
  XtFree( stupidstr );
}

void enterName( widget, closure, call_data )
Widget widget;
caddr_t closure, call_data;
{
  char *str;

  str = XmTextGetString( name );
  addToUsers( str );
  XtFree( str );

  XmTextSetString( name, NULL );
}

int getEvents( who, where )
char *who;
event **where;
{
  char **filsys;
  char filename[100], fallback[100], error[120];
  char *start;
  int cont = 1, e;
  struct stat info;

  MuSetWaitCursor( mainShell );

  e = HES_ER_OK;
  filsys = hes_resolve( who, "filsys" );
  if ( filsys == NULL )
    e = hes_error();
  if ( e != HES_ER_OK )
    switch ( e )
      {
      case 1:
	sprintf( error,
		"User %s not found in Hesiod database.\n", who );
	strcat( errors, error );
	cont = 0;
	break;
      default:
	sprintf( error, "Error %d from hes_resolve for user %s\n", e, who );
	strcat( errors, error );
	cont = 0;
	break;
      }
  else
    {
      sprintf( fallback, "%s", 1 + rindex( filsys[0], ' ') );
      if ( !strncmp( filsys[0], "AFS", 3 ) )
	{
#ifdef verbose
	  fprintf( stdout, "User %s has AFS homedir; not attaching\n", who );
#endif
	  start = index( filsys[0], ' ' ) + 1;

	  sprintf( filename, "%.*s", index( start, ' ' ) - start, start );
	}
      else
	strcpy( filename, fallback );
      
      if ( -1 == stat( filename, &info ) )
	switch( errno )
	  {
	  case ENOENT:
#ifdef verbose
	    fprintf( stdout, "Attempting to attach %s... ", who );
	    fflush( stdout );
#endif

	    sprintf( error, "attach -q %s", who );
	    e = system( error );
	    strcpy( filename, fallback ); /* for afs falling to attach */
	    switch( e )
	      {
	      case 0:
#ifdef verbose
		fprintf( stdout, "attached.\n" );
#endif
		break;
	      default:
		if ( -1 == stat( filename, &info ) )
		  {
		    sprintf( error, "Unsuccessful attach of %s", user );
		    cont = 0;
		  }
		else
		  {
		    sprintf( error,
			    "Attach returned error %d for %s; continuing\n",
			    e, user );
		    strcat( errors, error );
		  }
		break;
	      }
	    break;
	  case EACCES:
	    break;
/*
	    sprintf( error, "No permission to read directory %s\n", filename );
	    strcat( errors, error );
	    cont = 0;
	    break;
*/
	  default:
	    sprintf( error, "Error %d statting %s\n", errno, filename );
	    strcat( errors, error );
	    cont = 0;
	    break;
	  }

      if ( cont )
	{
	  strcat( filename, "/.tc" );
#ifdef verbose
	  fprintf( stdout, "Reading %s\n", filename );
#endif
	  if ( !addFileToEvents( filename, where ) )
	    {
	      cont = 0;
	      switch( errno )
		{
		case ENOENT:
		  sprintf( error, "%s has no calendar file.\n", who );
		  strcat( errors, error );
		  break;
		case EACCES:
		  sprintf( error,"No permission to read %s's calendar file.\n",
			   who );
		  strcat( errors, error );
		  break;
		default:
		  sprintf( error, "Error %d opening %s\n", errno, filename );
		  strcat( errors, error );
		  break;
		}
	    }
	}
    }
  MuSetStandardCursor( mainShell );
  return cont;
}

deletef()
{
  char *ridiculous;
  int *tbl, num, i;

  XmListGetSelectedPositions( userList, &tbl, &num );

  if ( num )
    {
      i = listmap[tbl[0]];

      /* this is ridiculous... I hate motif. */

      ridiculous = XmStringCreateLtoR( usernames[i],
				       XmSTRING_DEFAULT_CHARSET );
      XmListDeleteItem( userList, ridiculous );
      XtFree( ridiculous );

      /* s.b. common flush code */
      XtFree( usernames[i] );

      if ( userevents[i] != NULL )
	freeEvents( &userevents[i] );

      if ( busies[i] != NULL )
	{
	  XtFree( busies[i] );
	  busies[i] = NULL;
	}

      for ( i = listmap[tbl[0]]; i < numusers-1; i++ )
	{
	  usernames[i] = usernames[i+1];
	  busies[i] = busies[i+1];
	  userevents[i] = userevents[i+1];
	  ignore[i] = ignore[i+1];
	}

      for ( i = 0; i < numusers; i++ )
	if ( listmap[i] > listmap[tbl[0]] )
	  listmap[i] -= 1;

      for ( i = tbl[0]; i < numusers; i++ )
	listmap[i] = listmap[i+1];

      numusers--;
    }
  XtFree( tbl );
  updateGraph();
}

void updateGraph()
{
  int *tbl, num, i, cont, which;
  char error[120];
  int day, div;

  for ( day = 0; day < DAYS; day++ )
    for ( div = 0; div < HOURS*HDIVISIONS; div++ )
      busytimes.when[day][div] = 0;

  XmListGetSelectedPositions( userList, &tbl, &num );

  errors[0] = '\0';
  for ( i = 0; i < num; i++ )
    {
      which = listmap[tbl[i]];
      cont = 1;

      if ( ignore[which] == 0 && busies[which] == NULL )
	{
	  cont = getEvents( usernames[which], &userevents[which] );
	  
	  if ( cont )
	    {
	      if ( userevents[which] == NULL )
		{
		  sprintf( error, /* bug - will get printed over and over */
			  "User %s has an empty calendar file.\n",
			  usernames[which] );
		  strcat( errors, error );
		  cont = 0;
		}
	      else
		busies[which] = (busy (*)) XtMalloc( sizeof(busy) );
	    }
	  else
	    ignore[which] = 1;
	}

      if ( cont && !ignore[which] )
	{
	  eventsToBusy( userevents[which], busies[which],
		        sched.month, sched.day, sched.year );

	  for ( day = 0; day < DAYS; day++ )
	    for ( div = 0; div < HOURS*HDIVISIONS; div++ )
	      if ( busies[which]->when[day][div] != 0 )
		busytimes.when[day][div]++;
	}
    }
  if ( errors[0] != '\0' )
    MyError( 1, errors );

  redraw();
      
  XtFree( tbl );
}

void listSel( widget, closure, info )
Widget widget;
caddr_t closure;
XmListCallbackStruct *info;
{
  updateGraph();
}
