#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <des.h>
#include <krb.h>
#include "../server/network.h"
#include "../server/services.h"

#ifdef sun
extern char *sys_errlist;
#endif

#define instance "entropy"

char instrealm[80];
char errortext[200];
int msgcount;
connection *sender;
int inprogress;
int pflag = 0;

struct areq {
  int req;
  KTEXT_ST k;
} a;

CREDENTIALS cred;
des_key_schedule sched;
MSG_DAT msgdat;

void readmutauth();

void deadsock(c)
connection *c;
{
  fprintf(stdout, "socket died with error %d.\n", c->error);
  inprogress = -1;
}

char users[] = "cfields\nvanharen\n";
char data[] = "this is the\ntest application data\n:)\n";

void dump(ptr, len)
char *ptr;
int len;
{
  int i;

  for (i = 0; i < len; i++)
    putc(ptr[i], stdout);

  putc('\n', stdout);
  fflush(stdout);
}

void getget(c)
connection *c;
{
  int i;
  buffer *b;

  b = (buffer *)c->r_buffer;

  fprintf(stdout, "msgget returns (%d, %d)\n",
	  ntohl(b->srv_err), ntohl(b->sub_err));

  if (msgcount == -1) /* then first block; inited from tryget */
    {
      if (ntohl(b->srv_err) == ERR_MSG_RECV_SUCCESS)
	msgcount = ntohl(b->sub_err);
      else
	{
	  sprintf(errortext, "Server error returning transactions: %d\n%s\n",
		  ntohl(b->srv_err),
		  sys_errlist[ntohl(b->sub_err)]);
	  MyError(0, errortext);
	  inprogress = -1;

	  c->s_read = NULL;
	  free(c->r_buffer);
	  c->r_buffer = NULL;
	  return;
	}

      if (msgcount == 0)
	{
	  MyError(0, "No new transactions.");
	  inprogress = -1;

	  c->s_read = NULL;
	  free(c->r_buffer);
	  c->r_buffer = NULL;
	  return;
	}
    }
  else
    {
      msgcount--;
      if (msgcount == 0)
	{
	  c->s_read = NULL;
	  inprogress = 0;
	}
    }

  if (ERR_MSG_RECV_SUCCESS == ntohl(b->srv_err))
    for (i = 0; i < c->r_len - sizeof(*b); i++)
      fputc(b->data[i], stdout);
  fflush(stdout);  

  free(c->r_buffer);
  c->r_buffer = NULL;
}

void tryget(c)
connection *c;
{
  msgcount = -1;

  a.req = htonl(REQ_MSG_RECV);

  c->w_buffer = (char *)&a;
  c->w_len = sizeof(a.req);

  gag(c);
  c->s_write = NULL;
  c->s_read = getget;
}

int getAppointments()
{
  sender->s_dead = deadsock;
  sender->s_write = tryget;
  ungag(sender);

  inprogress = 1;
  while (inprogress == 1)
    {
      i_process(10);
      if (errtext != "\0")
	fprintf(stdout, errtext);
    }

  return inprogress;
}

void getsend(c)
connection *c;
{
  buffer *b;

  b = (buffer *)c->r_buffer;

  fprintf(stdout, "msgsend returns (%d, %d)\n",
	  ntohl(b->srv_err), ntohl(b->sub_err));

  free(c->r_buffer);
  c->r_buffer = NULL;
  c->s_read = NULL;
  c->s_write = tryget;
  ungag(c);
}

void trysend(c)
connection *c;
{
  int l;

  a.req = htonl(REQ_MSG_SEND);

  l = sprintf(&a.k, users) + 1;
  l += sprintf((char *)(&a.k) + l, data);

  c->w_buffer = (char *)&a;
  c->w_len = l + 4;

  gag(c);
  c->s_write = NULL;
  c->s_read = getsend;
}

void getgroups(c)
connection *c;
{
  int i;
  buffer *b;

  b = (buffer *)c->r_buffer;

  fprintf(stdout, "getgroups returns (%d, %d)\n",
	  ntohl(b->srv_err), ntohl(b->sub_err));

  if (ERR_GROUPS_ALL_SUCCESS == ntohl(b->srv_err))
    for (i = 0; i < c->r_len - sizeof(*b); i++)
      fputc(b->data[i], stdout);
  fflush(stdout);

  free(c->r_buffer);
  c->r_buffer = NULL;
  c->s_read = NULL;

  c->s_write = trysend;
  ungag(c);
}

void trygroups(c)
connection *c;
{
  a.req = htonl(REQ_GROUPS_ALL);
  c->w_buffer = (char *)&a;
  c->w_len = 4;

  gag(c);
  c->s_write = NULL;
  c->s_read = getgroups;
}

void writeauth(c)
connection *c;
{
  int k_err;

  strcpy(instrealm, krb_realmofhost(instance));
  k_err = krb_mk_req(&a.k, "rcmd", instance, instrealm, 5);

  if (k_err != KSUCCESS)
    {
      sprintf(errortext, "%s\n", krb_err_txt[k_err]);
      MyError(0, errortext);
      inprogress = -1;
      return;
    }

  a.req = htonl(REQ_AUTHENTICATE);
  c->w_buffer = (char *)&a;
  c->w_len = a.k.length + sizeof(a.k.length) + sizeof(a.req);

  gag(c);
  c->s_write = NULL;
  c->s_read = readmutauth;
}

void readmutauth(c)
connection *c;
{
  int k_err;

  if (ntohl(*(int *)(c->r_buffer)) == ERR_AUTHENTICATE_SUCCESS)
    {
      k_err = krb_get_cred("rcmd", instance, instrealm, &cred);

      if (k_err != KSUCCESS)
	{
	  sprintf(errortext,
		  "Mutual authentication failed on client side:\n%s\n",
		  krb_err_txt[k_err]);
	  MyError(0, errortext);
	  inprogress = -1;

	  c->s_read = NULL;
	  free(c->r_buffer);
	  return;
	}

      des_key_sched(cred.session, sched);

      k_err = krb_rd_priv(c->r_buffer + 2 * sizeof(int),
			  c->r_len,
			  sched,
			  cred.session,
			  &c->peer,
			  &c->us,
			  &msgdat);

      if (k_err != KSUCCESS)
	{
	  sprintf(errortext,
		  "Mutual authentication failed on client side:\n%s\n",
		  krb_err_txt[k_err]);
	  MyError(0, errortext);
	  inprogress = -1;

	  c->s_read = NULL;
	  free(c->r_buffer);
	  return;
	}

      fprintf(stdout, "data: %d %d %d %d\n",
	      msgdat.app_data[0],
	      msgdat.app_data[1],
	      msgdat.app_data[2],
	      msgdat.app_data[3]);
    }
  else
    {
      sprintf(errortext, "Server side authentication failure: %d;\n%s\n",
	      ntohl(*(int *)(c->r_buffer)),
	      krb_err_txt[ntohl(*(int *)(c->r_buffer + sizeof(int)))]);
      MyError(0, errortext);
      inprogress = -1;

      c->s_read = NULL;
      free(c->r_buffer);
      return;
    }

  c->s_read = NULL;
  free(c->r_buffer);
  inprogress = 0;
}

void initServices()
{
  static int initialized = 0;

  if (!initialized)
    {
      i_initialize();
      sender = i_create();
      initialized = 1;
    }
}

int authenticate()
{
  sender->s_dead = deadsock;
  sender->s_write = writeauth;
  ungag(sender);

  inprogress = 1;
  while (inprogress == 1)
    {
      i_process(10);
      if (errtext != "\0")
	fprintf(stdout, errtext);
    }

  return inprogress;
}

int openSocket()
{
  initServices();

  sender->machine = instance;
  sender->s_dead = deadsock;

  switch(i_open(sender))
    {
    case I_WAITING:
      fprintf(stdout, "waiting...\n");
      while (sender->status == S_CONNECTING)
	{
	  i_process(10);
	  if (errtext[0] != '\0')
	    fprintf(stdout, errtext);
	}

      if (sender->status == S_CONNECTED)
	return 0;
      break;

    case I_SUCCESS:
      return 0;
      break;

    default:
      sprintf(errortext, "Failure contacting server:\n%s\n", errtext);
      MyError(0, errortext);
      return -1;
      break;
    }

  sprintf(errortext, "Error contacting server:\n%s\n",
	  sys_errlist[sender->error]); /* should range check */
  MyError(0, errortext);

  return -1;
}
