#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <des.h>
#include <krb.h>
#include "network.h"
#include "services.h"

#define instance "entropy"
char instrealm[80];
int msgcount;
connection *sender;

struct areq {
  int req;
  KTEXT_ST k;
} a;

CREDENTIALS cred;
des_key_schedule sched;
MSG_DAT msgdat;

void readmutauth();

void deadsock(c)
connection *c;
{
  fprintf(stdout, "socket died.\n");
  exit(0);
}

char users[] = "cfields\nvanharen\n";
char data[] = "this is the\ntest application data\n:)\n";

void dump(ptr, len)
char *ptr;
int len;
{
  int i;

  for (i = 0; i < len; i++)
    putc(ptr[i], stdout);

  putc('\n', stdout);
  fflush(stdout);
}

void getget(c)
connection *c;
{
  int i;
  buffer *b;

  b = (buffer *)c->r_buffer;
  fprintf(stdout, "msgget returns (%d, %d)\n",
	  ntohl(b->srv_err), ntohl(b->sub_err));
  if (msgcount == -1)
    {
      if (ntohl(b->srv_err) == ERR_MSG_RECV_SUCCESS)
	msgcount = ntohl(b->sub_err);
      else
	{
	  c->s_read = NULL;
	  free(c->r_buffer);
	  c->r_buffer = NULL;
	  return;
	}
    }
  else
    {
      msgcount--;
      if (msgcount == 0)
	c->s_read = NULL;
    }

  if (ERR_MSG_RECV_SUCCESS == ntohl(b->srv_err))
    for (i = 0; i < c->r_len - sizeof(*b); i++)
      fputc(b->data[i], stdout);
  fflush(stdout);  

  free(c->r_buffer);
  c->r_buffer = NULL;
}

void tryget(c)
connection *c;
{
  msgcount = -1;

  a.req = htonl(REQ_MSG_RECV);

  c->w_buffer = (char *)&a;
  c->w_len = sizeof(a.req);

  gag(c);
  c->s_write = NULL;
  c->s_read = getget;
}

void getsend(c)
connection *c;
{
  buffer *b;

  b = (buffer *)c->r_buffer;

  fprintf(stdout, "msgsend returns (%d, %d)\n",
	  ntohl(b->srv_err), ntohl(b->sub_err));

  free(c->r_buffer);
  c->r_buffer = NULL;
  c->s_read = NULL;
  c->s_write = tryget;
  ungag(c);
}

void trysend(c)
connection *c;
{
  int l;

  a.req = htonl(REQ_MSG_SEND);

  l = sprintf(&a.k, users) + 1;
  l += sprintf((char *)(&a.k) + l, data);

  c->w_buffer = (char *)&a;
  c->w_len = l + 4;

  gag(c);
  c->s_write = NULL;
  c->s_read = getsend;
}

void getgroups(c)
connection *c;
{
  int i;
  buffer *b;

  b = (buffer *)c->r_buffer;

  fprintf(stdout, "getgroups returns (%d, %d)\n",
	  ntohl(b->srv_err), ntohl(b->sub_err));

  if (ERR_GROUPS_ALL_SUCCESS == ntohl(b->srv_err))
    for (i = 0; i < c->r_len - sizeof(*b); i++)
      fputc(b->data[i], stdout);
  fflush(stdout);

  free(c->r_buffer);
  c->r_buffer = NULL;
  c->s_read = NULL;

  c->s_write = trysend;
  ungag(c);
}

void trygroups(c)
connection *c;
{
  a.req = htonl(REQ_GROUPS_ALL);
  c->w_buffer = (char *)&a;
  c->w_len = 4;

  gag(c);
  c->s_write = NULL;
  c->s_read = getgroups;
}

void getecho(c)
connection *c;
{
  fprintf(stdout, "echoed: %d %d %d %d\n",
	  c->r_buffer[4], c->r_buffer[5], c->r_buffer[6], c->r_buffer[7]);
  c->s_read = NULL;
  free(c->r_buffer);
  c->r_buffer = NULL;

  c->s_write = trygroups;
  ungag(c);
}

void tryecho(c)
connection *c;
{
  char *x;

  a.req = htonl(REQ_ECHO);
  x = (char *)&a + 4;
  c->w_buffer = x - 4;
  c->w_len = 8;

  *x++ = 1;
  *x++ = 2;
  *x++ = 3;
  *x++ = 4;

  gag(c);
  c->s_write = NULL;
  c->s_read = getecho;
}

void writeauth(c)
connection *c;
{
  int k_err;

  strcpy(instrealm, krb_realmofhost(instance));
  k_err = krb_mk_req(&a.k, "rcmd", instance, instrealm, 5);

  if (k_err != KSUCCESS)
    {
      fprintf(stderr, "%s\n", krb_err_txt[k_err]);
      exit(1);
    }

  a.req = htonl(REQ_AUTHENTICATE);
  c->w_buffer = (char *)&a;
  c->w_len = a.k.length + sizeof(a.k.length) + sizeof(a.req);

  gag(c);
  c->s_write = NULL;
  c->s_read = readmutauth;
}

void readmutauth(c)
connection *c;
{
  int k_err;

  if (ntohl(*(int *)(c->r_buffer)) == ERR_AUTHENTICATE_SUCCESS)
    {
      k_err = krb_get_cred("rcmd", instance, instrealm, &cred);

      if (k_err != KSUCCESS)
	{
	  fprintf(stderr, "Mutual authentication failed on client side:\n%s\n",
		  krb_err_txt[k_err]);
	  exit(1);
	}

      des_key_sched(cred.session, sched);

      k_err = krb_rd_priv(c->r_buffer + 2 * sizeof(int),
			  c->r_len,
			  sched,
			  cred.session,
			  &c->peer,
			  &c->us,
			  &msgdat);

      if (k_err != KSUCCESS)
	{
	  fprintf(stderr, "Mutual authentication failed on client side:\n%s\n",
		  krb_err_txt[k_err]);
	  exit(1);
	}

      fprintf(stdout, "data: %d %d %d %d\n",
	      msgdat.app_data[0],
	      msgdat.app_data[1],
	      msgdat.app_data[2],
	      msgdat.app_data[3]);
    }
  else
    fprintf(stderr, "Server side authentication failure: %d;\n%s\n",
	    ntohl(*(int *)(c->r_buffer)),
	    krb_err_txt[ntohl(*(int *)(c->r_buffer + sizeof(int)))]);

  c->s_read = NULL;
  c->s_write = tryecho;
  ungag(c);
  free(c->r_buffer);
}

main()
{
  i_initialize();
  sender = i_create();

  sender->s_write = writeauth;
  sender->s_dead = deadsock;
  sender->machine = instance;

  switch(i_open(sender))
    {
    case I_WAITING:
      fprintf(stdout, "waiting...\n");
    case I_SUCCESS:
      ungag(sender);

      while (1)
	{
	  i_process(10);
	  if (errtext != "\0")
	    fprintf(stdout, errtext);
	}
      break;
    default:
      fprintf(stdout, "open: %s\n", errtext);
      exit(0);
    }
}
