/*
 * Need to read over carefully...
 * Errors to fix:
 *   Need to trap SIGPIPE
 *   make sure i_kill makes nfound, readfds, writefds consistent
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "network.h"

#define CLOSE		/* see comments at start of i_open routine */
/* #define DEBUG */

#define MAX(a, b) (a < b) ? b : a

char errtext[200]; /* externed in network.h */

int numconnections;
fd_set connections, gagged;
connection *head;
int maxfd;
int s_err, s_err_size;

void i_initialize()
{
  head = NULL;
  maxfd = 0;
  numconnections = 0;
  FD_ZERO(&connections);
  FD_ZERO(&gagged);

#ifdef DEBUG
  fprintf(stderr, "i_initialize: done.\n");
#endif
}

/*
 * Create and initialize a connection record
 */
connection *i_create()
{
  connection *c;

  errtext[0] = '\0';

  c = (connection *) malloc(sizeof(connection));

  c->machine = NULL;             /* no default machine */
  c->port = 32767;               /* default to 32767 */
  c->block = 0;                  /* default nonblocking */
  c->context = 0;                /* application defined */
  c->s_read = NULL;
  c->s_write = NULL;
  c->s_new = NULL;
  c->s_dead = NULL;
  c->s_empty = NULL;

  c->s = -1;
  c->status = S_NOTCONNECTED;

  c->readpending = 0;
  c->r_buffer = NULL;
  c->r_len = 0;
  c->r_pos = 0;

  c->gagged = 1;
  c->writepending = 0;
  c->w_buffer = NULL;
  c->w_len = 0;
  c->w_pos = 0;

  c->next = head;
  head = c;

#ifdef DEBUG
  fprintf(stderr, "i_create: record created.\n");
#endif
  return c;
}

/*
 * Use gethostbyname to get information about requested host
 */
int i_gethostbyname(hp, name)
struct hostent **hp;
char *name;
{
  int retries = RETRIES;

  while (retries)
    {
      if (NULL == (*hp = gethostbyname(name)))
	{
	  switch(h_errno)
	    {
	    case HOST_NOT_FOUND:
	      sprintf(errtext, "gethostbyname: host %s unknown", name);
	      return I_FAILURE;
	    case TRY_AGAIN:
	      retries--;
	      if (retries == 0)
		{
		  sprintf(errtext,
			  "gethostbyname: %d retries failed for %s",
			  RETRIES, name);
		  return I_FAILURE;
		}
	      break;
	    default:
	      sprintf(errtext, "gethostbyname: error %d for %s",
		      h_errno, name);
	      return I_FAILURE;
	    }
	}
      retries = 0;
    }
  return I_SUCCESS;
}

int i_open(c)
connection *c;
{
  struct hostent *hp; /* a record of info about the host to connect to */
  struct sockaddr_in destination; /* address record for connectee */
  int sus = sizeof(c->us);

  bzero((char *)&destination, sizeof(destination));
  errtext[0] = '\0';

  /*
   * Get a socket to make a connection on
   * Due to problems getting undocumented return values from CONNECT,
   * if after creating the socket we *ever* pass it in as the argument
   * to some function, for any error we will close the socket and get
   * a fresh one.
   */
  if (c->s == -1)
    if (-1 == (c->s = socket(PF_INET, SOCK_STREAM, 0)))
      {
	sprintf(errtext, "could not open a socket: error %d", errno);
#ifdef DEBUG
	fprintf(stderr, "i_open: %s.\n", errtext);
#endif DEBUG
	return I_FAILURE;
      }

  
#ifdef DEBUG
  fprintf(stderr, "i_open: created socket %d.\n", c->s);
#endif

  maxfd = MAX(c->s, maxfd);

  if (i_gethostbyname(&hp, c->machine) == I_FAILURE)
    {
#ifdef DEBUG
      fprintf(stderr, "i_open: %s on socket %d.\n", errtext, c->s);
#endif
      return I_FAILURE;
    }

  /*
   * Fill in internet address record
   */
  destination.sin_family = AF_INET;

  /* should check h_addrtype, but right now is always AF_INET */
  bcopy(hp->h_addr, &destination.sin_addr, hp->h_length);

  c->peer = destination;
  destination.sin_port = htons(c->port);

  /* set this socket non-blocking if desired */
  if (c->block == 0)
    if (-1 == fcntl(c->s, F_SETFL, FNDELAY))
      {
	sprintf(errtext, "fcntl call for blocking failed");
#ifdef DEBUG
	fprintf(stderr, "i_open: %s on socket %d.\n", errtext, c->s);
#endif
#ifdef CLOSE
	close(c->s);
	c->s = -1;
#endif
	return I_FAILURE;
      }

  fcntl(c->s, F_SETFD, 1); /* I don't think this can actually fail */

  /* attempt connection */
  if (-1 == connect(c->s, &destination, sizeof(destination)))
    switch(errno)
      {
      case ETIMEDOUT:
	sprintf(errtext, "connection timed out");
#ifdef DEBUG
	fprintf(stderr, "i_open: %s on socket %d.\n", errtext, c->s);
#endif
#ifdef CLOSE
	close(c->s);
	c->s = -1;
#endif
	return I_TIMEDOUT;

      case ECONNREFUSED:
	sprintf(errtext, "connection refused");
#ifdef DEBUG
	fprintf(stderr, "i_open: %s on socket %d.\n", errtext, c->s);
#endif
#ifdef CLOSE
	close(c->s);
	c->s = -1;
#endif
	return I_REFUSED;

      case EINPROGRESS:
	c->status = S_CONNECTING;
	numconnections++;

	/* on failure, this happens */
	FD_SET(c->s, &connections);

	/*
	 * on success, this happens
	 * note that this doesn't change the gag state, but makes sure
	 * we will be woken for success. the wakeup will restore flags
	 * to the gag state we really wanted.
	 */
	FD_SET(c->s, &gagged);

	sprintf(errtext, "connection attempt in progress");
#ifdef DEBUG
	fprintf(stderr, "i_open: %s on socket %d.\n", errtext, c->s);
#endif
	return I_WAITING;

      default:
	sprintf(errtext, "connection failure; error %d", errno );
#ifdef DEBUG
	fprintf(stderr, "i_open: %s on socket %d.\n", errtext, c->s);
#endif
#ifdef CLOSE
	close(c->s);
	c->s = -1;
#endif
	return I_FAILURE;
      }

#ifdef DEBUG
  fprintf(stderr, "i_open: successful connection to %s on socket %d.\n",
	  c->machine, c->s);
#endif

  bzero((char *) &c->us, sus);
  getsockname(c->s, &c->us, &sus);

  c->status = S_CONNECTED;
  numconnections++;
  FD_SET(c->s, &connections);
  return I_SUCCESS;
}

/*
 * Here's the deal: ungag the connection when you have something to say,
 * and your write procedure will be called as soon as the socket is ready
 * for output.
 * Once your write procedure sets up the proper pointers, if there isn't
 * going to be anything to say afterward, call gag. It will go ahead and
 * gag the socket after all of the writing completes.
 * If you want to gag at some other time for some strange reason, call
 * both gag and setgag.
 * Newly created sockets are always ga-mfff.
 */
void gag(c)
connection *c;
{
  c->gagged = 1;
}

void ungag(c)
connection *c;
{
  if (c->status == S_CONNECTED || c->status == S_CONNECTING)
    {
      FD_SET(c->s, &gagged);
      c->gagged = 0;
    }
}

void setgag(c)
connection *c;
{
  switch(c->gagged)
    {
    case 0:
      if (c->status == S_CONNECTED || c->status == S_CONNECTING)
	FD_SET(c->s, &gagged);
      break;

    case 1:
      FD_CLR(c->s, &gagged);
      break;
    }
}

int i_listen(c)
connection *c;
{
#define NAMELEN 50
  struct hostent *hp; /* a record of info about us */
  char name[NAMELEN];
  struct sockaddr_in destination; /* address record for us */
  int sus = sizeof(c->us);

  errtext[0] = '\0';

  /*
   * Get a socket to accept connections on
   */
  if (c->s == -1)
    if (-1 == (c->s = socket(PF_INET, SOCK_STREAM, 0)))
      {
	sprintf(errtext, "could not open a socket: error %d", errno);
#ifdef DEBUG
	fprintf(stderr, "i_listen: %s.\n", errtext);
#endif DEBUG
	return I_FAILURE;
      }

#ifdef DEBUG
  fprintf(stderr, "i_listen: created socket %d.\n", c->s);
#endif

  maxfd = MAX(c->s, maxfd);

  gethostname(name, NAMELEN); /* we can't lose! :) :) */
  if (i_gethostbyname(&hp, name) == I_FAILURE)
    {
#ifdef DEBUG
      fprintf(stderr, "i_listen: %s on socket %d.\n", errtext, c->s);
#endif
      return I_FAILURE;
    }

  /*
   * Fill in internet address record
   */
  destination.sin_family = AF_INET;
  destination.sin_port = htons(c->port);
  destination.sin_addr.s_addr = INADDR_ANY;

  /* set this socket non-blocking if desired */
  if (c->block == 0)
    if (-1 == fcntl(c->s, F_SETFL, FNDELAY))
      {
	sprintf(errtext, "fcntl call for blocking failed");
#ifdef DEBUG
	fprintf(stderr, "i_listen: %s on socket %d.\n", errtext, c->s);
#endif
	return I_FAILURE;
      }

  fcntl(c->s, F_SETFD, 1); /* I don't think this can actually fail */

  if (-1 == bind(c->s, &destination, sizeof(destination)))
    {
      sprintf(errtext, "bind failed with error %d", errno);
#ifdef DEBUG
	fprintf(stderr, "i_listen: %s on socket %d.\n", errtext, c->s);
#endif
      return I_FAILURE;
    }

  bzero((char *) &c->us, sus);
  c->us.sin_addr.s_addr = gethostid();

  if (-1 == listen(c->s, MAXQUEUE))
    {
      sprintf(errtext, "error %d in listen", errno);
#ifdef DEBUG
	fprintf(stderr, "i_listen: %s on socket %d.\n", errtext, c->s);
#endif
      return I_FAILURE;
    }
 
#ifdef DEBUG
  fprintf(stderr, "i_listen: listen initiated on socket %d.\n", c->s);
#endif

  c->status = S_LISTENING;
  numconnections++;
  FD_SET(c->s, &connections);
  return I_SUCCESS;
}

int i_write(c)
connection *c;
{
  int netl, sent;

  errtext[0] = '\0';

  if (!c->writepending)
    {
#ifdef DEBUG
      fprintf(stderr, "i_write: beginning write on socket %d.\n", c->s);
#endif
      c->writepending++;
      c->w_pos = -sizeof(netl);
    }

  if (c->w_pos < 0)
    {
      netl = htonl(c->w_len);
      sent = write(c->s, (&netl) + sizeof(netl) + c->w_pos, - c->w_pos);
      if (sent == -1)
	{
	  /* here's a bug... */
	  sprintf(errtext, "uh-oh: write socket %d, error %d", c->s, errno);
#ifdef DEBUG
	  fprintf(stderr, "i_write: %s. Time to debug!\n", errtext);
#endif
	  return T_INCOMPLETE; /* this should do s_dead() */
	}
      c->w_pos += sent;
      if (c->w_pos < 0)
	return T_INCOMPLETE;
    }

  sent = write(c->s, c->w_buffer + c->w_pos, c->w_len - c->w_pos);
  if (sent == -1)
    {
      sprintf(errtext, "uh-oh: write socket %d, error %d", c->s, errno);
#ifdef DEBUG
      fprintf(stderr, "i_write: %s. Time to debug! (the sequel)\n", errtext);
#endif
      return T_INCOMPLETE; /* this should do s_dead() */
    }
  c->w_pos += sent;
  if (c->w_pos < c->w_len)
    return T_INCOMPLETE;

#ifdef DEBUG
      fprintf(stderr, "i_write: write completed on socket %d (%d bytes).\n",
	      c->s, c->w_len);
#endif

  c->writepending = 0;

  if (c->s_empty != NULL) /* let application free its data if it wishes */
    c->s_empty(c);

  c->w_pos = 0;
  c->w_len = 0;
  c->w_buffer = NULL;
  setgag(c);

  return T_COMPLETE;
}

void i_kill(c, e)
connection *c;
int e;
{
  int cl_err;

/*  fprintf(stdout, "got: %d\n", e); */

#ifdef DEBUG
  fprintf(stderr, "i_kill: socket %d died, error %d.\n", c->s, e);
#endif

  cl_err = close(c->s);

#ifdef DEBUG
  if (cl_err == -1)
    fprintf(stderr, "i_kill: close error %d (ouch!).\n", errno);
#endif

  FD_CLR(c->s, &connections);
  FD_CLR(c->s, &gagged);

  c->s = -1;
  c->readpending = 0;
  c->status = S_NOTCONNECTED;
  c->error = e;

  if (c->s_dead != NULL)
    c->s_dead(c);

/* Application can free it whenever it's ready, but we'll forget about it.
  if (c->r_buffer != NULL)
    free(c->r_buffer);
*/
  c->r_buffer = NULL;
}

void i_close(c)
connection *c;
{
  int cl_err;

  /* should check to see if it's connected first... */
  cl_err = close(c->s);

#ifdef DEBUG
  if (cl_err == -1)
    fprintf(stderr, "i_close: close error %d (ouch!).\n", errno);
#endif

  FD_CLR(c->s, &connections);
  FD_CLR(c->s, &gagged);

  c->s = -1;
  c->readpending = 0;
  c->status = S_NOTCONNECTED;
  c->error = 0;

  c->r_buffer = NULL;
}

/*
 * i_destroy()
 *   doesn't exist yet. :)
 */

void i_read(c)
connection *c;
{
  int received;

  errtext[0] = '\0';

  if (!c->readpending)
    {
#ifdef DEBUG
      fprintf(stderr, "i_read: beginning read on socket %d.\n", c->s);
#endif
      c->readpending++;
      c->r_pos = -sizeof(c->r_len);
    }

  if (c->r_pos < 0)
    {
      received = read(c->s, &c->r_len + sizeof(c->r_len) +
		      c->r_pos, - c->r_pos);
      if (received == -1)
	{
	  sprintf(errtext, "uh-oh: read socket %d, error %d", c->s, errno);
#ifdef DEBUG
	  fprintf(stderr, "i_read: %s. Time to debug!\n", errtext);
#endif
	  return; /* no errors should be possible here */
	}

      if (received == 0) /* end of file; connection closed */
	{
	  s_err = 0; s_err_size = sizeof(s_err);
	  getsockopt(c->s, SOL_SOCKET, SO_ERROR, &s_err, &s_err_size);

	  i_kill(c, s_err);
	  return;
	}

      c->r_pos += received;
      if (c->r_pos == 0)
	{
	  c->r_len = ntohl(c->r_len);
	  if (c->r_len != 0)
	    c->r_buffer = (char *)malloc(c->r_len);
	}
      else
	return; /* haven't read entire length longword yet */
    }

  received = read(c->s, c->r_buffer + c->r_pos, c->r_len - c->r_pos);
  if (received == -1)
    {
      if (errno == EWOULDBLOCK)
	return; /* reasonable if we just got 4 bytes above... */
      sprintf(errtext, "uh-oh: read socket %d, error %d", c->s, errno);
#ifdef DEBUG
      fprintf(stderr, "i_read: %s. Time to debug! (again)\n", errtext);
#endif
      return; /* this should do s_dead() (i_kill) */
    }

  if (received == 0)
    {
      s_err = 0; s_err_size = sizeof(s_err);
      getsockopt(c->s, SOL_SOCKET, SO_ERROR, &s_err, &s_err_size);

      /* bug in error structure here */
      i_kill(c, s_err);
      return;
    }

  c->r_pos += received;
  if (c->r_pos < c->r_len)
    return;

#ifdef DEBUG
      fprintf(stderr, "i_read: read completed on socket %d (%d bytes).\n",
	      c->s, c->r_len);
#endif

  if (c->s_read != NULL)
    c->s_read(c);
  else
    free(c->r_buffer); /* if s_read, it will do it */

  c->readpending = 0;
  c->r_pos = 0;
  c->r_len = 0;
  c->r_buffer = NULL;
  return;
}

void pr(f, fds)
char *f;
fd_set *fds;
{
  int i, start = 0;

  for (i = 0; i < 256; i++)
    if (FD_ISSET(i, fds))
      {
	if (!start)
	  {
	    start = 1;
	    fprintf(stderr, f);
	  }
	else
	  fprintf(stderr, ",");
	fprintf(stderr, "%d", i);
      }
  if (start)
    fprintf(stderr, "\n");
}

int i_process(timeout)
int timeout;
{
  static int nfound = 0;
  struct timeval tv;
  static fd_set readfds, writefds, exceptfds;
  struct sockaddr_in peer; /* address record for connectee */
  int peerlen;
  connection *i, *tmp, *tmpagain;
  int tmps;

  errtext[0] = '\0';

  tv.tv_sec = timeout;  /* 0 for just a poll */
  tv.tv_usec = 0;

  /* if there's nothing we know we need to do, ask... */
  if ( nfound == 0 )
    {
      readfds = exceptfds = connections;
      writefds = gagged;
      nfound = select( maxfd+1, &readfds, &writefds, &exceptfds, &tv );
      if (nfound == -1)
	{
	  fprintf(stderr, "i_process: select returned error %d.\n", errno);
	  exit(1); /* toast! this is wrong. not just wrong, damn wrong. */
	}
    }

  if (nfound == 0)
    {
#ifdef DEBUG
      fprintf(stderr, "i_process: select returned nothing to do.\n");
#endif
      return -1;
    }

#ifdef NOTPOSSIBLYDEFINEDISIT
  fprintf(stderr, "i_process: something to do.\n");
  pr("read: ", &readfds);
  pr("write: ", &writefds);
  pr("except: ", &exceptfds);
#endif

  /*
   * do any reading that wants to be done
   */
  for (i = head; i != NULL; i = i->next)
    if (FD_ISSET(i->s, &readfds))
      switch(i->status)
	{
	  /*
	   * Case for accepting a new connection
	   */
	case S_LISTENING:
	  peerlen = sizeof(peer);
	  tmps = accept(i->s, &peer, &peerlen);
	  if (tmps == -1)
	    {
	      sprintf(errtext, "uh-oh: accept %d", errno);
	      nfound--;
#ifdef DEBUG
	      fprintf(stderr,
		      "i_process: bad news for socket %d: error %d.\n",
		      i->s, errno);
#endif
	      return nfound; /* bug! */
	    }

#ifdef DEBUG
	  fprintf(stderr, "i_process: new connection on socket %d.\n", tmps);
#endif
	  tmp = i_create(); /* structure for our new friend */
	  tmpagain = tmp->next; /* let's not screw up the list */

	  *tmp = *i; /* breaks the list */

	  tmp->next = tmpagain; /* fixes it again */
	  tmp->s = tmps;
	  tmp->peer = peer;

	  maxfd = MAX(tmps, maxfd);

	  tmp->status = S_CONNECTED;
	  if (tmp->s_new != NULL)
	    tmp->s_new(tmp);
	  FD_SET(tmps, &connections);
	  nfound--;
	  FD_CLR(i->s, &readfds);
	  return nfound;

	  /*
	   * A socket gets selected for read before it is connected
	   * to indicate connection failed.
	   */
	case S_CONNECTING:
	  s_err_size = sizeof(s_err);
	  getsockopt(i->s, SOL_SOCKET, SO_ERROR, &s_err, &s_err_size);
#ifdef DEBUG
	  fprintf(stderr, "i_process: error %d on connecting socket %d.\n",
		  s_err, i->s);
#endif
	  FD_CLR(i->s, &readfds);

	  if (FD_ISSET(i->s, &writefds)) /* god knows why... damn unix. */
	    {
	      FD_CLR(i->s, &writefds);
	      nfound--;
	    }

	  i_kill(i, s_err);
	  nfound--;
	  return nfound;

	  /*
	   * The standard case...
	   */
	case S_CONNECTED:
	  i_read(i);
	  nfound--;
	  FD_CLR(i->s, &readfds);
	  return nfound;

	  /*
	   * The bizarre case...
	   */
	case S_NOTCONNECTED:
	  fprintf(stderr, "i_process: big-time bug.\n");
	  nfound--;
	  FD_CLR(i->s, &readfds);
	  return nfound;
	}

  /*
   * do any writing that is ready to be done
   */
  for (i = head; i != NULL; i = i->next)
    if (FD_ISSET(i->s, &writefds))
      {
	/* should case this out as above. but hey, I'm lazy. */

	/*
	 * A connecting socket gets selected for write to indicate
	 * a successful connection.
	 */
	if (i->status == S_CONNECTING)
	  {
	    int sus = sizeof(i->us);
#ifdef DEBUG
	    fprintf(stderr, "i_process: connection completed on socket %d.\n",
		    i->s);
#endif
	    bzero((char *) &i->us, sus);
	    getsockname(i->s, &i->us, &sus);

	    i->status = S_CONNECTED;
	    if (i->s_new != NULL)
	      i->s_new(i);
	    nfound--;
	    FD_CLR(i->s, &writefds);
	    setgag(i);
	    return nfound;
	  }

	if (i->writepending)
	  i_write(i);
	else
	  if (i->s_write != NULL)
	    {
	      i->s_write(i);
	      i_write(i);
	    }

	nfound--;
	FD_CLR(i->s, &writefds);
	return nfound;
      }

  /* it was for exceptfds; s.b. only oob data, which we don't use. ignore. */
  /* bug: should just not select for it */
#ifdef DEBUG
  for (i = head; i != NULL; i = i->next)
    if (FD_ISSET(i->s, &exceptfds))
      fprintf(stderr, "i_process: exceptional condition for %d.\n", i->s);
  if (nfound)
    sleep(2);
#endif

  sprintf( errtext, "except: %d", nfound );

  nfound = 0;
  return nfound;
}
