extern char errtext[];

#define RETRIES 3
#define MAXQUEUE 5

#define S_CONNECTED 0
#define S_NOTCONNECTED 1
#define S_CONNECTING 2
#define S_LISTENING 3   /* for listening sockets */

#define T_INCOMPLETE 0
#define T_COMPLETE 1

#define I_SUCCESS 0
#define I_FAILURE 1    /* generic failure */
#define I_TIMEDOUT 2
#define I_REFUSED 3
#define I_WAITING 4    /* connection not made immediately */

typedef void (*func)();

/*
 * s_write() gets called anytime the socket is ready for writing
 * s_read() gets called for each full record that has arrived on the socket
 * s_new() is called for any newly created socket, however it was done
 * s_dead() is called for non-locally broken sockets
 */
typedef struct _connection {

  /* Initializable by application */
  int context;     /* application's context number */
  char *machine;   /* name of machine connect(ed/ing) to */
  int port;        /* port number of machine */
  struct sockaddr_in us;   /* address record for us */
  struct sockaddr_in peer; /* address record for peer */
  int block;       /* whether this socket should block or not */
  func s_read;     /* procedure to call when socket is ready to be read */
  func s_write;    /* procedure to call when socket is ready to be written */
  func s_new;      /* procedure to call on new connections */
  func s_dead;     /* procedure to call for dead sockets */
  func s_empty;    /* procedure to call when write is done with data */

  /* application should not write */
  int s;           /* file descriptor of created socket */
  int status;      /* status of this connection */
  int error;       /* error number if connection died */

  int gagged;        /* wake up application when write ready? */
  int writepending;  /* this buffer is only partially written */
  char *w_buffer;    /* buffer */
  int w_len;         /* length of data in buffer to be sent */
  int w_pos;         /* position to continue writing in buffer */

  int readpending;   /* this buffer is only partially read */
  char *r_buffer;    /* buffer */
  int r_len;         /* length of data in buffer to be sent */
  int r_pos;         /* position to continue writing in buffer */

  struct _connection *next;
} connection;

extern void i_initialize();
extern connection *i_create();
extern int i_open();
extern int i_listen();
extern int i_process();
