#include <ctype.h>
#include <stdio.h>
#include "structs.h"

/* #define STAND */

/* for now */
#ifdef STAND
#define XtMalloc malloc
#define XtFree free
#define filename "structs.dat"
#endif

int ampm = 1;
int assume_day = 1;

/*
 * The purpose of matchbits is to return a boolean (though it
 * really doesn't) indicating whether bit num is set in bits.
 * This is used to help check if a simple date is included in
 * a complicated one.
 */
int matchBits( num, bits )
int num, bits;
{
  return (bits & (1<<(num-1)));
}

/*
 * Turn a string expression into a set of bits (in an int)
 * The bits would probably be appropriately be abstracted into
 * a type set.
 * 
 * expression:                      bits:
 *
 *      *                           all
 *     a-b                          bits a to b set
 *                                  if b is less than a, then
 *                                  bits from b to max, and
 *                                  0 to a are set.
 *      a                           bit a is set
 *
 * All expressions but * must be enclosed in {}, and multiple
 * expressions may be separated by commas. Thus
 * {a-b, c-d, e, f, g-h} is a valid expression, as is *.
 */
int expToBits( e, max )
char *e;
int max;
{
  int i, j, n, first = 0, range; /* shutup! */
  int rtn = 0;

  if ( e[0] == '*' )
    return (int)~0;
  if ( e[0] != '{' )
    return ( 1<<(atoi(e)-1) );

  i = 1; n = 0; range = 0;
  while ( i < strlen( e ) )
    {
      while ( e[i] == ' ' || e[i] == '-' || e[i] == ',' || e[i] == '}' )
	i++;

      while ( e[i] != ' ' && e[i] != '-' && e[i] != ',' && e[i] != '}' )
	n = 10 * n + e[i++] - '0';

      switch( e[i] )
	{
	case ' ':
	  break;
	case '-':
	  range = 1;
	  first = n;
	  n = 0;
	  break;
	case ',':
	case '}':
	  if ( range )
	    {
	      range = 0;
	      if ( n < first )
		n += max; /* this with mod to wrap */
	      for ( j = first; j <= n; j++ )
		rtn |= (1<<((j-1)%max));
	    }
	  else
	    rtn |= (1<<(n-1));
	  n = 0;
	  if ( e[i] == '}' )
	    i = strlen( e );
	  break;
	default: /* buggy stuff */
	  fprintf( stderr, "expToBits: %c\n", e[i] );
	  break;
	}
    }
  return rtn;
}

/*
 * Converts a day string expression to its bits
 */
int dayToRep( w )
char *w;
{
  return expToBits( w, 7 ) & wildDays;
}

/*
 * Converts a month string expression to its bits
 */
int monthToRep( m )
char *m;
{
  return expToBits( m, 12 ) & wildMonths;
}

/*
 * Converts a date string expression to its bits
 */
int dateToRep( d )
char *d;
{
  return expToBits( d, 31 ) & wildDates;
}

/*
 * Converts a year string expression (limited) to an int.
 * No ranges are allowed in its representation, except for wild.
 */
int yearToRep( y )
char *y;
{
  if ( y[0] == '*' )
    return wildYear;
  else
    return atoi( y );
}

int parse(what, where)
char what[], *where;
{
  int i = 0;

  while (what[i] && isspace(what[i]))
    i++;

  while (what[i] && !isspace(what[i]))
    *where++ = what[i++];

  *where = '\0';

  return i;
}

/*
 * If *wm == NULL, malloc for it and initialize it
 */
void checkinit(wm)
constraints **wm;
{
  if (*wm == NULL)
    {
      *wm = (constraints *)malloc(sizeof(constraints));

      (*wm)->lowerbound.month = wildInt;
      (*wm)->lowerbound.day = wildInt;
      (*wm)->lowerbound.year = wildInt;
      (*wm)->upperbound.month = wildInt;
      (*wm)->upperbound.day = wildInt;
      (*wm)->upperbound.year = wildInt;

      (*wm)->days = wildDays;
      (*wm)->dates = wildDates;
      (*wm)->months = wildMonths;
      (*wm)->year = wildYear;
      (*wm)->more = NULL;
    }
}

/*
 * add an initialized empty event to the head of the events
 * list and return it
 */
event *newEvent(events)
event **events;
{
  event *new;

  new = (event *)XtMalloc(sizeof(event));

  new->when = NULL;
  new->info = (description *)XtMalloc(sizeof(description));
  new->next = *events;

  new->info->desc = NULL;

  new->info->starttime.hour = wildHour;
  new->info->starttime.minute = wildMinute;
  new->info->starttime.second = wildSecond;

  new->info->endtime = new->info->starttime;

  new->info->reminder = wildInt;
  new->info->scheduled = wildInt;
  new->info->transaction = wildInt;
  new->info->source = NULL;
  new->info->recipients = NULL;
  new->info->status = wildInt;
  new->info->accepted = wildInt;

  *events = new;

  return new;
}

#define M_BASE 0
#define M_TEXT 1

static constraints **wm;
static int mode, linesleft, used;
static char Description[DESCMAX];

void startEvent(new)
event *new;
{
  mode = M_BASE;
  wm = &new->when;
}

void endEvent(new)
event *new;
{
  if (mode == M_TEXT) /* aborted prematurely; salvage */
    {
      if (used)
	{
	  new->info->desc = (char *)XtMalloc(used + 1);
	  bcopy(Description, new->info->desc, used + 1);
	  mode = M_BASE;
	}
    }
}

int eventLine(new, t)
event *new;
char *t;
{
  char m[10], d[100], y[10], flag[30], param[30];
  int hs, ms, ss, he, me, se;
  int i;

  if (mode == M_TEXT)
    {
      strncat(&Description[used], t, DESCMAX-used);
      used += strlen(&Description[used]);
      linesleft--;
      if (linesleft == 0)
	{
	  new->info->desc = (char *)XtMalloc(used + 1);
	  bcopy(Description, new->info->desc, used + 1);
	  mode = M_BASE;
	  return 0; /* we're done with this record */
	}
      else
	return 1;
    }

  switch(t[0])
    {
    case '{':            /* start of environment */
      break;
    case '}':            /* end of environment */
      if (*wm)
	wm = &((*wm)->more);
      break;
    case '[':
    case ']':
      break;
    case 'w':
      checkinit(wm);
      (*wm)->days = dayToRep(&t[1]);
      break;
    case 'd':
      checkinit(wm);

      if ((*wm)->dates != wildDates ||
	  (*wm)->months != wildMonths ||
	  (*wm)->year != wildYear)
	{
	  wm = &((*wm)->more);
	  checkinit(wm);
	}

      sscanf(&t[1], " %[^/]/%[^/]/%[^ ]", m, d, y);
      (*wm)->months = monthToRep( m );
      (*wm)->dates = dateToRep( d );
      (*wm)->year = yearToRep( y ); /* must restructure */
      break;
    case 's':
      sscanf(&t[1], " %d:%d:%d", &hs, &ms, &ss);
      new->info->starttime.hour = hs;
      new->info->starttime.minute = ms;
      new->info->starttime.second = ss;
      break;
    case 'e':
      sscanf(&t[1], " %d:%d:%d", &he, &me, &se);
      new->info->endtime.hour = he;
      new->info->endtime.minute = me;
      new->info->endtime.second = se;
      break;
    case 'p':
      i = parse(&t[1], flag) - 1;
      if (!strcmp(flag, "highlight"))
	new->info->highlight = atoi(t[i]);
      else
	if (!strcmp(flag, "scheduled"))
	  new->info->scheduled = atoi(t[i]);
	else
	  if (!strcmp(flag, "transaction"))
	    new->info->transaction = atoi(t[i]);
	  else
	    if (!strcmp(flag, "accepted"))
	      new->info->accepted = atoi(t[i]);
	    else
	      if (!strcmp(flag, "status"))
		new->info->status = atoi(t[i]);
	      else
		if (!strcmp(flag, "source"))
		  {
		    (void)parse(&t[i], param);
		    new->info->source = XtNewString(param);
		  }
      break;
    case 'r':
      break;
    case 't':
      mode = M_TEXT;
      linesleft = atoi(&t[1]);
      used = 0;
      Description[0] = '\0';
      if (linesleft == 0)
	{
	  mode = M_BASE; /* in case I'm stupid later... */
	  return 0; /* no more lines for this event */
	}
      break;
    }
  return 1; /* more lines should still come to this event */
}

/*
 * Takes a file and adds it to an events list.
 * An events list is a fully general list of all events the program
 * knows about.
 *
 * version 1 file format:
 *
 *    {days} {mm}/{dd}/yyyyv* hh:mm:ss hh:mm:ss a b c
 *    # Header text line (for single line display)
 *    # description text line 1...
 *    # description text line n...
 * 
 * any of {n} may be formed {n1,n2-n3,n5}
 * a b c doesn't exist in version 0; c was used in version 1 (for "daylike")
 *
 * * * * * * * * * * * * *
 * version 2 file format:
 *
 * [ mm/dd/yyyy         closed lower bound on days
 * ] mm/dd/yyyy         closed upper bound on days
 * w {days}             days of week mask
 * d {mm}/{dd}/yyyyv*   date mask
 * s hh:mm:ss           start time of event
 * e hh:mm:ss           end time of event
 * {                    not supported yet...
 * }                     "      "      "
 * p <param> <value>    set a special parameter for this message
 * r n                  recipient list follows (n of them, including status)
 * t n                  n lines of text follow; ends an entry
 *
 * All fields are optional except t. :)
 *
 * General params:
 *
 *   scheduled          boolean, shows up in busy time?
 *   reminder           what kind of reminders for this?
 *   highlight          highlight days for this entry? (reread suggestion...)
 *
 * Appointment specific params:
 *
 *   source             string; username of source
 *   transaction        integer; transaction id from server
 *   accepted           let's remember whether we said yes, no, or nothing
 *   status             status of this transaction; final, cancelled, tentative
 *   
 * For the spec, the time specifying fields may be entered multiple times.
 * They will be unioned and intersected in mysterious ways (the {}'s add
 * capability here) to be able to specify lots of interesting things.
 */
int addFileToEvents( file, events )
char *file;
event **events;
{
  FILE *f;
  event *new;
  constraints *w, **wm;
  char wk[10], m[10], d[100], y[10];
  char l, t[DESCMAX];
  int used, nowhere;
  int hs, ms, ss, he, me, se;
  int err, version = 0;

  f = fopen( file, "r" );

  if ( f == NULL ) /* needs better stuff here */
    {
      *events = NULL;
      return 0;
    }

  l = getc(f); /* i think this dies on null files *bug* */
  ungetc(l, f);

  if (l == '_')
    {
      fscanf(f, "_version %d\n", &version);
      l = getc(f);
      ungetc(l, f);
    }

  switch(version)
    {
    case 0:
    case 1:
      while (!feof(f))
	{
	  new = (event *)XtMalloc(sizeof(event));

	  new->when = NULL;
	  new->info = (description *)XtMalloc(sizeof(description));
	  new->next = *events;
	  new->info->desc = NULL;
	  *events = new;

	  wm = &new->when;

	  while (l != EOF && l != '#')
	    {
	      /* for versions 0 and 1, there will always be constraints */
	      w = (constraints *)XtMalloc(sizeof(constraints));
	      *wm = w;
	      w->more = NULL;
	      wm = &w->more;

	      switch(version)
		{
		case 0:
		  fscanf(f, "%[^ ] %[^/]/%[^/]/%[^ ] %d:%d:%d %d:%d:%d\n",
			 wk, m, d, y, &hs, &ms, &ss, &he, &me, &se);
		  break;
		case 1:
		  fscanf(f,
		  "%[^ ] %[^/]/%[^/]/%[^ ] %d:%d:%d %d:%d:%d %d %d %d\n",
			 wk, m, d, y, &hs, &ms, &ss, &he, &me, &se,
			 &nowhere, &nowhere, &nowhere);
		}

	      w->days = dayToRep( wk );
	      w->months = monthToRep( m );
	      w->dates = dateToRep( d );
	      w->year = yearToRep( y ); /* must restructure */

	      if (version == 1 && nowhere)
		{
		  new->info->starttime.hour =
		    new->info->starttime.minute =
		      new->info->starttime.second = wildInt;
		  new->info->endtime = new->info->starttime;
		}
	      else
		{
		  new->info->starttime.hour = hs; /* bogus */
		  new->info->starttime.minute = ms;
		  new->info->starttime.second = ss;
		  new->info->endtime.hour = he;
		  new->info->endtime.minute = me;
		  new->info->endtime.second = se;
		}
	      l = getc( f );
	      ungetc( l, f );
	    }

	  w->lowerbound.month = wildInt; /* no such fields in 0 and 1... */
	  w->lowerbound.day = wildInt;
	  w->lowerbound.year = wildInt;

	  w->upperbound.month = wildInt;
	  w->upperbound.day = wildInt;
	  w->upperbound.year = wildInt;

	  new->info->reminder = wildInt;
	  new->info->scheduled = 1;
	  new->info->transaction = wildInt;
	  new->info->source = NULL;
	  new->info->recipients = NULL;
	  new->info->status = wildInt;
	  new->info->accepted = wildInt;

	  used = 0;

	  while (l != EOF && l == '#' && used != DESCMAX-1)
	    {
	      l = getc(f);
	      fgets(&t[used], DESCMAX-used, f); /* includes cr */
	      used = strlen(t);

	      l = getc(f);
	      ungetc(l, f);
	    }

	  /* bug; put in code to skip characters over DESCMAX */

	  if ( used != 0 )
	    {
	      new->info->desc = (char *) XtMalloc( used+1 );
	      bcopy( t, new->info->desc, used+1 );
	    }
	}
      break;
    case 2:
      while (!feof(f))
	{
	  err = (int)fgets(t, DESCMAX, f);
	  if (err == 0)
	    break;

	  new = newEvent(events);

	  startEvent(new);
	  while (err != 0 && eventLine(new, t))
	    err = (int)fgets(t, DESCMAX, f);

	  endEvent(new);
	}
      break; /* case 2 */
    }

  fclose( f );
  return 1;
}

void all( bits, str )
int bits;
char *str;
{
  char work[10];
  char bld[100];
  int i, first;
  int multi = 0, range = 0, started = 0;

  sprintf( bld, "" );

  for ( i = 1; i < 33; i++ ) /* highest bit will be zero. */
    {
      /*
       * If we're not in a range, and this bit is set, start a range
       */
      if ( (!range) && (bits & 1) )
	{
	  range = 1;
	  first = i;
	}
      else
	/*
	 * If we are in a range, and this bit is clear, end the range
	 */
	if ( range && !(bits & 1) )
	  {
	    /*
	     * If this is not the first range, tack a comma on
	     */
	    if ( !started )
	      started = 1;
	    else
	      strcat( bld, "," );
	    
	    range = 0;
	    if ( i == first+1 )
	      sprintf( work, "%d", first );
	    else
	      {
		multi = 1;
		sprintf( work, "%d-%d", first, i-1 );
	      }
	    strcat( bld, work );
	  }
      bits = bits >> 1;
    }
  if ( multi || index( bld, ',' ) )
    sprintf( str, "{%s}", bld );
  else
    strcpy( str, bld );
}

int first( bits )
int bits;
{
  int i;
  for ( i = 0; i < 31; i++ )
    {
      if ( bits & 1 )
	return i + 1;
      bits = bits >> 1;
    }
}

int dumpEventsToFile( events, file )
char *file;
event *events;
{
  FILE *f;
  char *s, ms[100], ds[100];
  constraints *w;
  int usebraces, numlines, extra;

  f = fopen( file, "w" );
  if ( f == NULL )
    return 0;

  fprintf( f, "_version 2\n" );
  while (events != NULL)
    {
      w = events->when;
      if (w)
	{
	  if (w->more)
	    usebraces = 1;
	  else
	    usebraces = 0;
	}

      while ( w != NULL )
	{
	  all( w->months, ms );
	  all( w->dates, ds );

/* for now, we don't need to be so smart...
	  if (usebraces)
	    fprintf(f, "{\n");
*/

	  fprintf(f, "d %s/%s/%d\n", ms, ds, w->year); /* year wrong */

/*
	  if (usebraces)
	    fprintf(f, "}\n");
*/

	  w = w->more;
	}

      if (events->info->starttime.hour != wildHour ||
	  events->info->starttime.minute != wildMinute ||
	  events->info->starttime.second != wildSecond)
	fprintf(f, "s %d:%d:%d\n",
		events->info->starttime.hour,     /* all wrong for wild case */
		events->info->starttime.minute,   /* but ok, since only all can */
		events->info->starttime.second);  /* be wild */

      if (events->info->endtime.hour != wildHour ||
	  events->info->endtime.minute != wildMinute ||
	  events->info->endtime.second != wildSecond)
	fprintf(f, "e %d:%d:%d\n",
		events->info->endtime.hour,
		events->info->endtime.minute,
		events->info->endtime.second);

      extra = numlines = 0;
      s = events->info->desc;
      if (s)
	{
	  while (*s != '\0')
	    if (*s++ == '\n')
	      numlines++;

	  if (*events->info->desc != '\0')
	    if (s[-1] != '\n')
	      {
		extra = 1;
		numlines++;
	      }
	}

      if (events->info->scheduled != wildInt)
	fprintf(f, "p scheduled %d\n", events->info->scheduled);
      if (events->info->highlight != wildInt)
	fprintf(f, "p highlight %d\n", events->info->highlight);

      if (events->info->transaction != wildInt)
	fprintf(f, "p transaction %d\n", events->info->transaction);
      if (events->info->accepted != wildInt)
	fprintf(f, "p accepted %d\n", events->info->accepted);
      if (events->info->status != wildInt)
	fprintf(f, "p status %d\n", events->info->status);
      if (events->info->source != NULL)
	fprintf(f, "p source %s\n", events->info->source);

      fprintf(f, "t %d\n", numlines);
      fprintf(f, "%s", events->info->desc);
      if (extra)
	fputc('\n', f);

      events = events->next;
    }
  fclose( f );
  return 1;
}

deleteFromEvents( e, what )
event **e;
description *what;
{
  event *last = NULL, *current;
  constraints *w1, *w2;

  current = *e;

  while ( current != NULL )
    {
      if ( current->info == what )
	{
	  w1 = current->when;
	  while ( w1 != NULL )
	    {
	      w2 = w1;
	      w1 = w1->more;
	      XtFree( w2 );
	    }
	  XtFree( current->info->desc );
	  XtFree( current->info );
	  if ( last == NULL )
	    *e = current->next;
	  else
	    last->next = current->next;
	  XtFree( current );
	  current = NULL;
	}
      else
	{
	  last = current;
	  current = current->next;
	}
    }
}

/*
 * We make our own copy of all of the information
 */
void addToEvents( e, sels, numsels, ofni )
event **e;
CalendarSelection *(sels[]);
int numsels;
description *ofni;
{
  int i, j;
  event *new;
  constraints *ptr;

  new = (event *)XtMalloc(sizeof(event));

  new->next = *e;

  new->info = (description *)XtMalloc(sizeof(description));
  *(new->info) = *(ofni);
  new->info->desc = (char *)XtMalloc(strlen(ofni->desc) + 1); /* bug */
  bcopy(ofni->desc, new->info->desc, strlen(ofni->desc) + 1);

  ptr = (constraints *)XtMalloc(sizeof(constraints));
  new->when = ptr; /* bugs for null selections... */
  *e = new;

  for (i = 0; i < numsels; i++)
    {
      if (i > 0)
	{
	  ptr->more = (constraints *)XtMalloc(sizeof(constraints));
	  ptr = ptr->more;
	}

      ptr->more = NULL;
      ptr->days = wildDays;

      ptr->months = 1 << ( (sels[i])[0].month - 1);
      ptr->year = (sels[i])[0].year;
      ptr->dates = 0;

      for ( j = 0; (sels[i])[j].day != 0; j++ )
	ptr->dates |= 1 << ( (sels[i])[j].day - 1);
    }
}

/*
 * Used by cvtEventToSimple to copy a simple event into
 * a list it is building. Enters it sorted.
 */
void addSimpToList( simp, list )
simpleEvent *simp, **list;
{
  simpleEvent *temp;
  simpleEvent *ptr;
  simpleEvent **lastPtr;

  temp = (simpleEvent *) XtMalloc( sizeof( simpleEvent ) );
  temp->info = simp->info;

  if ( *list == NULL )
    {
      *list = temp;
      temp->next = NULL;
      return;
    }

  ptr = *list;
  lastPtr = list;

  while ( ptr != NULL ) /* Yuuuuuuuuuuuck!!! */
    if (
	ptr->info->starttime.hour != wildHour
	&&
	(
	 (
	  (ptr->info->starttime.hour != wildHour && /* for daylike stuff */
	   temp->info->starttime.hour == wildHour)
	  ) ||
	 (
	  ((ptr->info->starttime.hour * 60) +  ptr->info->starttime.minute) >
	  ((temp->info->starttime.hour * 60) + temp->info->starttime.minute)
	  )
	 )
	)
      {
	temp->next = ptr;
	*lastPtr = temp;
	return;
      }
    else
      {
	lastPtr = &(ptr->next);
	ptr = ptr->next;
      }

  if ( ptr == NULL )
    {
      temp->next = NULL;
      *lastPtr = temp;
    }
}

static int py=0, pm=0, p[42];

getp( year, month )
int year, month;
{
  if ( year != py || month != pm )
    {
      py = year;
      pm = month;
      cal( pm, py, p );
    }
}

/*
 * return day of week (1 - 7) (Sun - Sat)
 */
int dayof( year, month, day )
{
  int i;

  getp( year, month );

  for ( i = 0; i < 42; i++ )
    if ( p[i] == day )
      return ( i%7 + 1 );
}

/*
 * return a *calendarselection array containing all the days that d occurs
 * this is a hack until I write the more general stuff, and so lacks
 * many useful capabilities
 */
getalldates( e, d, sels, numsels )
event *e;
description *d;
CalendarSelection *(sels[]);
int *numsels;
{
  int day, dw;
  constraints *when;
  int i = 0, j;
  int month, year;

  while ( e != NULL )
    {
      if ( d == e->info ) /* here we are! */
	{
	  when = e->when;
	  while ( when != NULL )
	    {
	      j = 0;
	      month = first( when->months );
	      year = when->year;

	      dw = dayof( year, month, 1 );
	      /* this is scrod in 1752 */
	      for ( day = 1; day < 32; day++ )
		{
		  /* shit... structure being blown away */
		  if ( sels[i] == NULL )
		    sels[i] = (CalendarSelection *)
		      XtMalloc( 32 * sizeof( CalendarSelection ) );

		  if ( matchBits( day, when->dates ) &&
		      matchBits( dw, when->days ))
		    {
		      (sels[i])[j].month = month;
		      (sels[i])[j].year = year;
		      (sels[i])[j].day = day;
		      j++;
		    }
		  dw = dw%7 + 1;
		}
	      (sels[i])[j].month = month;
	      (sels[i])[j].year = year;
	      (sels[i])[j].day = 0;
	      when = when->more;
	      i++;
	    }
	  e = NULL;
	}
      else
	e = e->next;
    }
  *numsels = i;
}

/*
 * nothing calls this... it's an old routine that will probably go soon
 *
 * return a calendarselection array containing the days that d occurs
 * in a given month
 */
getdates( e, d, year, month, s )
event *e;
description *d;
int month, year;
CalendarSelection s[];
{
  int day, dw;
  constraints *when;
  int i = 0;

  while ( e != NULL )
    {
      if ( d == e->info )
	{
	  when = e->when;
	  while ( when != NULL )
	    {
	      if ( when->year == year || when->year == 0 )
		if ( matchBits( month, when->months ) )
		  {
		    dw = dayof( year, month, 1 );
		    /* this is scrod in 1752 */
		    for ( day = 1; day < 32; day++ )
		      {
			if ( matchBits( day, when->dates ) &&
			    matchBits( dw, when->days ))
			  {
			    s[i].year = year;
			    s[i].month = month;
			    s[i].day = day;
			    i++;
			  }
			dw = dw%7 + 1;
		      }
		  }
	      when = when->more;
	    }
	  e = NULL;
	}
      else
	e = e->next;
    }
  s[i].month = month;
  s[i].day = 0;
  s[i].year = year;
}

/*
 * Return a list of simpleEvents containing all entries in the
 * event list e that match to month/day/year.
 * SimpleEvents are useful structures to higher level routines
 * that want to print the events for a day; events are in a
 * slightly difficult format to use, which is why this conversion
 * abstraction exists. The event structure format is also likely
 * to change. And already has...
 */
simpleEvent *cvtEventToSimple( e, year, month, day )
event *e;
int year, month, day;
{
simpleEvent *list, s;
constraints *when;

list = NULL;

while ( e != NULL )
  {
    when = e->when;
    while ( when != NULL )
      {
	if ( when->year == year || when->year == 0 )
	  if ( matchBits( month, when->months ) )
	    if ( matchBits( day, when->dates ) )
	      if ( matchBits( dayof( year, month, day ), when->days ) )
		{
		  s.info = e->info;
		  addSimpToList( &s, &list );
		}
	when = when->more;
      }
    e = e->next;
  }
return list;
}

/*
 * Free a list of simpleEvents that is not needed anymore.
 */
freeSimple( s )
simpleEvent *s;
{
simpleEvent *p;

p = s;
while ( p != NULL )
  {
    /* don't free description - something else points to it */
    XtFree( p );
    p = p->next; /* "contents undisturbed" */
  }
}

/*
 * Free a list of events
 */
void freeEvents( e )
event **e;
{
  event *p, *q;
  constraints *d, *f;

  p = *e;

  while ( p != NULL )
    {
      XtFree( p->info->desc ); /* free description text */
      XtFree( p->info ); /* free description */

      d = p->when;
      while ( d != NULL ) /* free date list */
	{
	  f = d;
	  d = d->more;
	  XtFree( f );
	}
      q = p;
      p = p->next;
      XtFree( q );
    }
  *e = NULL;
}

/*
 * Debugging routine
 */
void printSimp( list )
simpleEvent *list;
{
simpleEvent *here;

here = list;
while ( here != NULL )
  {
    fprintf( stdout, "%d %d %d\n",
	     here->info->starttime.hour,
	     here->info->starttime.minute,
	     here->info->starttime.second );
    if ( here->info->desc != NULL )
      fprintf( stdout, "%s\n", here->info->desc );
    here = here->next;
  }
}

void timeToStr( t, str )
time t;
char *str;
{
  char ap[] = {' ','m','\0'};
  int h;

  if ( !ampm )
    sprintf( str, "%02d:%02d",
	     t.hour, t.minute );
  else
    {
      h = t.hour;
      if ( h < 12 )
	{
	  if ( h == 0 )
	    h = 12;
	  ap[0] = 'a';
	}
      else
	{
	  ap[0] = 'p';
	  if ( h > 12 )
	    h -= 12;
	}
      sprintf( str, "%2d:%02d %s",
	       h, t.minute, ap );
    }
}

/*
 * Given an integer, return its suffix (5th, 22nd...)
 * you didn't see this...
 */
char *suffix( d )
int d;
{
  static char *suff[] = { "st", "nd", "rd", "th",};

  if ( d > 9 && d < 21 )
    return suff[3];
  if ( ((d-1)%10) > 2 )
    return suff[3];
  return suff[(d%10) - 1];
}

char    *smon1[]= {
        "January", "February", "March", "April",
        "May", "June", "July", "August",
        "September", "October", "November", "December",
};

char *month( i )
int i;
{
  return smon1[i - 1];
}

void getTimeNum( s, n )
char **s;
int *n;
{
  char *e;
  int q;

  while ( isspace( **s ) ) (*s)++;
  e = *s;

  q = 0;
  while ( isdigit( *e ) ) { q = 10*q + ( *e - '0' ); e++; }
  if ( e == *s )
    q = -1;
  *n = q;
  *s = e;

  while ( isspace( **s ) ) (*s)++;
  if ( **s == ':' )
    (*s)++;
}

void strToTime( s, t )
char *s;
time *t;
{
  int spec = 0;

  getTimeNum( &s, &t->hour );
  getTimeNum( &s, &t->minute );
  getTimeNum( &s, &t->second );

  if ( t->minute == -1 )
    t->minute = 0;
  if ( t->second == -1 )
    t->second = 0;

  if ( (islower( *s ) ? toupper( *s ) : *s) == 'P' )
    {
      if ( t->hour < 12 )
	t->hour += 12;
      spec = 1;
    }
  if ( (islower( *s ) ? toupper( *s ) : *s) == 'A' )
    {
      if ( t->hour > 11 )
	t->hour -= 12;
      spec = 1; /* if they said 15:00AM, tough */
    }

  if ( !spec && assume_day )
    if ( t->hour < 8 )
      t->hour += 12;
}

/*
#ifdef STAND
main()
{
  event *events;
  simpleEvent *simp;

  events = NULL;

  addFileToEvents( filename, &events );
  simp = cvtEventToSimple( events, 1573, 9, 2 );
  printSimp( simp );
}
#endif
*/
