#include "Calendar.h"

#define wildInt -1
#define wildHour -1               /* meaning none specified */
#define wildMinute -1             /* no wild times implemented */
#define wildSecond -1

#define wildDays 127              /* 2^7 - 1 */
#define wildDates 2147483647      /* 2^31 - 1 */
#define wildMonths 4095           /* 2^12 - 1 */
#define wildYear -1               /* 1^2 - 2 */

#define DESCMAX 2000 /* max chars in event description */

typedef struct _time
{
  int hour, minute, second;
} time;

typedef struct _date
{
  int month, day, year;
} date;

typedef struct _constraints       /* time constraints of record */
{
  date lowerbound, upperbound;
  int days;                       /* bitmap of days of week */
  int dates;                      /* bitmap of dates */
  int months;                     /* bitmap of months */
  int year;                       /* not a bitmap :) */
  struct _constraints *more;      /* in case the spec is complex */
} constraints;

#define STAT_TENTATIVE 0
#define STAT_FINALIZED 1
#define STAT_CANCELLED 2

#define ACC_NORESPONSE 0
#define ACC_YES 1
#define ACC_NO 2

#define TYPE_SIMPLE -1
#define TYPE_SERVER 0
#define TYPE_REPLY 1

typedef struct _description       /* descriptive elements of record */
{
  int type;			  /* the type of this entry. determines the
				     interpretation of some later records */
  int limbo;			  /* has user dealt with this entry?
				     not relevant for TYPE_SIMPLE */
  int transaction;                /* transaction number from server; useful
				     only for TYPE_SERVER */
  int reference;		  /* reference transaction number (for
				     TYPE_SERVER (supercedes) or TYPE_REPLY) */
  char *source;                   /* username who initiated this transaction */
  char **recipients;              /* list of recipients if applicable */

  int status;                     /* status of this transaction... see STAT_ */

  time starttime, endtime;        /* when it is (not for TYPE_REPLY) */
  char *desc;                     /* text description for TYPE_SIMPLE and
				     TYPE_SERVER; comment if TYPE_REPLY */

  int accepted;                   /* what we replied if this is NEW or
				     UPDATE. reply if REPLY. see ACC_ */

  /* the following cannot come from the server; they are user parameters */
  int reminder;                   /* reminder level */
  int scheduled;                  /* boolean; shows up on appointments? */
  int highlight;
} description;

typedef struct _event
{
  constraints *when;              /* time specifier; null would be everywhen */
  description *info;
  struct _event *next;
} event;

/*
 * A simpleEvent exists for the purposes of a non-gross interface to
 * the application. The application need not know about the details
 * of the underlying event representations. It can just ask that all
 * of the events in the database that occur in a given month be placed
 * into a linked list of simpleEvents, which contains the information
 * the application really needs to care about.
 */
typedef struct _simpleEvent
{
  description *info;
  struct _simpleEvent *next;
} simpleEvent;

extern int ampm;
extern int assume_day;
extern simpleEvent *cvtEventToSimple();
extern int addFileToEvents();
extern void freeEvents();
extern void timeToStr();
extern void strToTime();
extern void getTimeNum();
